from mako.template import Template


class CSVFormatterV1:
    """
    This class is responsible for formatting items to strings
    representing individual rows in a CSV file, and for exporting
    collections of items to CSV files.
    """

    def __init__(self, fmt_dict, sep=","):
        """
        Initialise CSV formatter.

        Parameters
        ----------
        fmt_dict: dict
            Dictionary which maps column names to mako templating
            strings. The column names are used as the headers when
            exporting a collection of items to a CSV file (but are
            ignored when formatting single items).
        sep: string, default ','
            Delimiter between individual fields in each row.
        """
        self.fmt_dict = fmt_dict
        self.sep = sep
        self.default_header_line = "#" + sep.join(fmt_dict.keys()) + "\n"
        self.fmt_template = Template(sep.join(fmt_dict.values()))
        self.line_separator = "\n"

    def format_item(self, item):
        """
        Return a string representation of `item`.

        Parameters
        ----------
        item: namedtuple
            The item to be formatted.
        """
        return self.fmt_template.render(**item._asdict())

    def get_header_line(self, header):
        if isinstance(header, str):
            return header + self.line_separator
        elif header is True:
            return self.default_header_line
        elif header is False:
            return ""
        else:
            raise ValueError("Invalid 'header' argument (must be boolean or string): '{}'".format(header))

    def to_csv(self, items, path_or_buf=None, *, header=True):
        """
        Export items to CSV file.

        Parameters
        ----------
        items: iterable
            The items to export. Each item corresponds to
            one row in the CSV file.
        path_or_buf: string or file handle, default None
            File path or object, if None is provided the
            result is returned as a string.
        header: boolean or list of string, default True
            Write out a header line with column names. If
            a list of string is given it is assumed to be
            aliases for the column names
        """
        csv = self.get_header_line(header)
        for item in items:
            csv += self.format_item(item) + self.line_separator

        if path_or_buf is None:
            return csv
        else:
            if isinstance(path_or_buf, str):
                with open(path_or_buf, 'w') as f:
                    f.write(csv)
            else:
                path_or_buf.write(csv)



from .context import tohu
from tohu.generators import Integer, Float, Sequential, TupleGenerator
from tohu.utils import First, Second, Nth, BufferedTuple, Split, Zip


class DummyTupleGen(TupleGenerator):
    """
    Dummy generator class which generates a
    sequence of tuples with increasing counter.

    Example:

    >>> g = DummyTupleGen("Foo_", "Bar_", "Baz_", digits=1)
    >>> next(g)
    ("Foo_01", "Bar_01", "Baz_01")
    >>> next(g)
    ("Foo_02", "Bar_02", "Baz_02")
    >>> next(g)
    ("Foo_03", "Bar_03", "Baz_03")

    """
    def __init__(self, *prefixes):
        self.cnt = 0
        self.prefixes = prefixes
        self.tuple_len = len(prefixes)

    def __next__(self):
        self.cnt += 1
        return tuple(prefix + str(self.cnt) for prefix in self.prefixes)


class TestBufferedTuple:

    def test_tuple_elements_can_be_extracted_in_arbitrary_order(self):
        """
        Test that extracting individual tuple elements in
        arbitrary order produces the expected elements.
        """

        pair_gen = DummyTupleGen("Foo", "Bar")
        b = BufferedTuple(pair_gen)

        assert b.next_nth(0) == "Foo1"
        assert b.next_nth(1) == "Bar1"
        assert b.next_nth(0) == "Foo2"
        assert b.next_nth(0) == "Foo3"
        assert b.next_nth(1) == "Bar2"
        assert b.next_nth(0) == "Foo4"
        assert b.next_nth(1) == "Bar3"
        assert b.next_nth(1) == "Bar4"
        assert b.next_nth(1) == "Bar5"
        assert b.next_nth(1) == "Bar6"
        assert b.next_nth(0) == "Foo5"


class TestUtils:

    def test_extract_first_element_from_tuple_generator(self):
        """
        Test that `First` produces a generator yielding the first element in each tuple.
        """
        pairs = DummyTupleGen("a", "b")
        triples = DummyTupleGen("c", "d", "e")
        g = First(pairs)
        h = First(triples)
        assert list(g.generate(5)) == ["a1", "a2", "a3", "a4", "a5"]
        assert list(h.generate(3)) == ["c1", "c2", "c3"]

    def test_extract_second_element_from_tuple_generator(self):
        """
        Test that `Second` produces a generator yielding the second element in each tuple.
        """
        pairs = DummyTupleGen("a", "b")
        triples = DummyTupleGen("c", "d", "e")
        g = Second(pairs)
        h = Second(triples)
        assert list(g.generate(5)) == ["b1", "b2", "b3", "b4", "b5"]
        assert list(h.generate(3)) == ["d1", "d2", "d3"]

    def test_extract_nth_element_from_tuple_generator(self):
        """
        Test that `Nth` produces a generator yielding the n-th element in each tuple.
        """
        pairs = DummyTupleGen("a", "b")
        triples = DummyTupleGen("c", "d", "e")
        g = Nth(pairs, 0)
        h = Nth(triples, 2)
        assert list(g.generate(4)) == ["a1", "a2", "a3", "a4"]
        assert list(h.generate(3)) == ["e1", "e2", "e3"]

    def test_split_pairs(self):
        """
        Test that `Split` produces a generators yielding the individual elements
        in each tuple produced by the original generator
        """
        pairs = DummyTupleGen("Foo", "Quux")

        g, h = Split(pairs)

        assert next(g) == "Foo1"
        assert next(h) == "Quux1"
        assert next(g) == "Foo2"
        assert next(g) == "Foo3"
        assert next(g) == "Foo4"
        assert next(h) == "Quux2"
        assert next(h) == "Quux3"
        assert next(g) == "Foo5"
        assert next(h) == "Quux4"
        assert next(h) == "Quux5"
        assert next(h) == "Quux6"
        assert next(h) == "Quux7"
        assert next(g) == "Foo6"

    def test_split_triples(self):
        """
        Test that `Split` produces a generators yielding the individual elements
        in each tuple produced by the original generator
        """
        triples = DummyTupleGen("Foo", "Bar", "Quux")

        g1, g2, g3 = Split(triples)

        assert next(g1) == "Foo1"
        assert next(g2) == "Bar1"
        assert next(g3) == "Quux1"
        assert next(g1) == "Foo2"
        assert next(g1) == "Foo3"
        assert next(g1) == "Foo4"
        assert next(g3) == "Quux2"
        assert next(g1) == "Foo5"
        assert next(g2) == "Bar2"
        assert next(g2) == "Bar3"
        assert next(g3) == "Quux3"
        assert next(g3) == "Quux4"
        assert next(g3) == "Quux5"

    def test_zip(self):
        """

        """
        g1 = Sequential(prefix="Foo_", digits=3)
        g2 = Integer(0, 1000)
        g3 = Float(0, 1)

        g = Zip(g1, g2, g3, seed=99999)

        assert next(g) == ("Foo_001", 123, 0.12049047611259744)
        assert next(g) == ("Foo_002", 972, 0.3092850036925837)
        assert next(g) == ("Foo_003", 316, 0.9529198903583602)


from collections import namedtuple
import pytest
import textwrap
from .context import tohu
from tohu.csv_formatter import CSVFormatter


class TestCSVFormatter:

    @classmethod
    def setup(cls):
        """
        Create a few records of type `Foobar` to be used in tests.
        """
        Foobar = namedtuple("Foobar", ["aaa", "bbb", "ccc"])
        cls.records = [
            Foobar(aaa='foobar_01', bbb=8, ccc='4898FE19'),
            Foobar(aaa='foobar_02', bbb=160, ccc='5825D187'),
            Foobar(aaa='foobar_03', bbb=99, ccc='3648A436')
        ]

    def test_init_with_field_dict(self):
        """
        Test that CSVFormatter can be initialised with a dictionary of field names and their format strings.
        """
        fields = {
            'Column 1': 'a=${aaa}',
            'Column 2': 'b=${bbb}',
            'Column 3': 'c=${ccc}',
        }
        csv_formatter = CSVFormatter(fields=fields)
        csv = csv_formatter.format_records(self.records)

        csv_expected = textwrap.dedent("""\
            #Column 1,Column 2,Column 3
            a=foobar_01,b=8,c=4898FE19
            a=foobar_02,b=160,c=5825D187
            a=foobar_03,b=99,c=3648A436
            """)

        assert csv == csv_expected

    def test_init_with_field_dict_and_custom_field_separator(self):
        """
        Test that custom field_separator is taken into account.
        """
        fields = {
            'Column 1': 'a=${aaa}',
            'Column 2': 'b=${bbb}',
            'Column 3': 'c=${ccc}',
        }
        csv_formatter = CSVFormatter(fields=fields, sep=" || ")
        csv = csv_formatter.format_records(self.records)

        csv_expected = textwrap.dedent("""\
            #Column 1 || Column 2 || Column 3
            a=foobar_01 || b=8 || c=4898FE19
            a=foobar_02 || b=160 || c=5825D187
            a=foobar_03 || b=99 || c=3648A436
            """)

        assert csv == csv_expected

    def test_init_with_format_string(self):
        """
        Test that CSVFormatter can be initialised with a formatting string.
        """
        fmt_str = "a=${aaa}, b: ${bbb}, c has value ${ccc}"
        csv_formatter = CSVFormatter(fmt_str=fmt_str)
        csv = csv_formatter.format_records(self.records)

        csv_expected = textwrap.dedent("""\
            a=foobar_01, b: 8, c has value 4898FE19
            a=foobar_02, b: 160, c has value 5825D187
            a=foobar_03, b: 99, c has value 3648A436
            """)

        assert csv == csv_expected

    def test_format_records_multiline_format_string(self):
        """
        Test that formatting a few records to CSV using a multiline format string produces the expected result.
        """

        header = "#Column_1,Column_2,LineNoOfRecord"
        fmt_str = "a=${aaa},b=${bbb},first_line\na=${aaa},c=${ccc},second_line"
        csv_formatter = CSVFormatter(fmt_str=fmt_str, header=header)
        csv = csv_formatter.format_records(self.records)

        csv_expected = textwrap.dedent("""\
            #Column_1,Column_2,LineNoOfRecord
            a=foobar_01,b=8,first_line
            a=foobar_01,c=4898FE19,second_line
            a=foobar_02,b=160,first_line
            a=foobar_02,c=5825D187,second_line
            a=foobar_03,b=99,first_line
            a=foobar_03,c=3648A436,second_line
            """)

        assert csv == csv_expected

    def test_write_csv_file(self, tmpdir):
        """
        Test that writing records to a CSV file produces the expected result.
        """
        filename = tmpdir.join("output.csv").strpath

        csv_formatter = CSVFormatter(fmt_str="${aaa},${bbb},${ccc}", header="# Custom header line")
        csv_formatter.write(filename, self.records)

        csv = open(filename).read()
        csv_expected = textwrap.dedent("""\
            # Custom header line
            foobar_01,8,4898FE19
            foobar_02,160,5825D187
            foobar_03,99,3648A436
            """)

        assert csv == csv_expected

    def test_subdirectories_are_created_if_necessary(self, tmpdir):
        """
        Test that
        """
        filename = tmpdir.join("foo/bar/output.csv").strpath

        assert not tmpdir.join("foo").exists()
        assert not tmpdir.join("foo/bar").exists()

        csv_formatter = CSVFormatter(fmt_str="${aaa},${bbb},${ccc}", header="# Custom header line")
        csv_formatter.write(filename, self.records)

        assert tmpdir.join("foo").exists()
        assert tmpdir.join("foo/bar").exists()

        csv = open(filename).read()
        csv_expected = textwrap.dedent("""\
            # Custom header line
            foobar_01,8,4898FE19
            foobar_02,160,5825D187
            foobar_03,99,3648A436
            """)

        assert csv == csv_expected



class TestCSVFormatterHeaderLine:
    """
    This class tests the various combinations of initialising
    a CSV formatter and the resulting header line of the
    generated CSV.
    """

    @classmethod
    def setup_class(cls):
        cls.fields = {
            "Column 1": "${}",
            "Column 2": "${}",
            "Column 3": "${}",
        }

    def test_init_with_field_dict_and_no_header_argument(self):
        formatter = CSVFormatter(fields=self.fields, header=None)
        assert formatter.format_records(records=[]) == "#Column 1,Column 2,Column 3\n"

    def test_init_with_field_dict_and_header_True(self):
        formatter = CSVFormatter(fields=self.fields, header=True)
        assert formatter.format_records(records=[]) == "#Column 1,Column 2,Column 3\n"

    def test_init_with_field_dict_and_header_False(self):
        formatter = CSVFormatter(fields=self.fields, header=False)
        assert formatter.format_records(records=[]) == ""

    def test_init_with_field_dict_and_custom_header(self):
        formatter = CSVFormatter(fields=self.fields, header="# This is a custom header line")
        assert formatter.format_records(records=[]) == "# This is a custom header line\n"

    def test_init_with_field_dict_and_custom_separator(self):
        formatter = CSVFormatter(fields=self.fields, sep=" || ", header=True)
        assert formatter.format_records(records=[]) == "#Column 1 || Column 2 || Column 3\n"

    def test_init_with_format_str_and_no_header_argument(self):
        formatter = CSVFormatter(fmt_str="", header=None)
        assert formatter.format_records(records=[]) == ""

    def test_init_with_format_str_and_header_False(self):
        formatter = CSVFormatter(fmt_str="", header=False)
        assert formatter.format_records(records=[]) == ""

    def test_init_with_format_str_and_custom_header(self):
        formatter = CSVFormatter(fmt_str="", header="# This is a custom header line")
        assert formatter.format_records(records=[]) == "# This is a custom header line\n"

    def test_init_with_format_str_and_header_True_raises_error(self):
        """
        This should raise an error because the header can't be automatically derived from the `fmt_str` argument.
        """
        with pytest.raises(ValueError):
            _ = CSVFormatter(fmt_str="", header=True)



import os
from mako.template import Template


class CSVFormatter:
    """
    This class formats an individual record or
    a sequence of records as CSV output.
    """

    def __init__(self, *, fields=None, fmt_str=None, sep=",", header=None):
        """
        Initialise CSV formatter.

        Either `fields` or `fmt_str` must be specified (but not both).

        If `fields` is given (which must be a dictionary of column names
        and associated format strings) then a header line is automatically
        created from the column names, i.e. the keys of the dictionary.
        However, this can be overriden by passing the `header` argument.

        Parameters
        ----------
        fmt_str: string
            Formatting string that specifies how an
            individual record should be formatted.

        sep: string
            Field separator. Default: ','

        header: string (optional)
            Header line to be used in the CSV output.
        """
        self.line_separator = "\n"
        self.sep = sep

        if fields is not None:
            if header is None or header is True:
                self.header_line = "#" + self.sep.join(fields.keys()) + self.line_separator
            elif header is False:
                self.header_line = ""
            elif isinstance(header, str):
                self.header_line = header + self.line_separator
            else:
                raise ValueError("Invalid 'header' argument: '{}' (must be string or boolean or None).".format(header))

            fmt_str = self.sep.join(fields.values())
            self.template = Template(fmt_str + self.line_separator)
        elif fmt_str is not None:
            if header is True:
                raise ValueError("Cannot automatically derive header. Please provide an explicit header string.")
            self.header_line = "" if (header is None or header is False) else (header + self.line_separator)
            self.template = Template(fmt_str + self.line_separator)
        else:
            raise ValueError("Either `fields` or `fmt_str` must be given.")

    def format_record(self, record):
        """
        Return formatted string representing the individual record.
        """
        return self.template.render(**record._asdict())

    def format_records(self, records):
        """
        Return CSV string of
        """
        s = self.header_line
        for r in records:
            s += self.template.render(**r._asdict())
        return s

    def write(self, filename, items):
        """
        Write items to file.
        """
        dirname = os.path.dirname(filename)

        if not os.path.exists(dirname):
            os.makedirs(dirname)

        with open(filename, 'w') as f:
            f.write(self.header_line)
            for item in items:
                f.write(self.format_record(item))


from collections import deque
from functools import partial
from .generators import BaseGenerator, GeolocationPair, TupleGenerator

__all__ = ['First', 'Second', 'Nth', 'Split', 'Geolocation']


class Nth(BaseGenerator):
    """
    Generator which allows to extract the n-th element from a tuple-producing generator.
    """

    def __init__(self, g, idx):
        self.g = g
        self.idx = idx

    def __next__(self):
        return next(self.g)[self.idx]

    def _spawn(self):
        return Nth(self.g._spawn(), self.idx)

    def reset(self, seed):
        self.g.reset(seed)


First = partial(Nth, idx=0)
Second = partial(Nth, idx=1)


class BufferedTuple(BaseGenerator):
    """
    Helper class which allows buffered extraction
    of items from a tuple generator.
    """

    def __init__(self, g, *, maxbuffer=10):
        """
        Parameters
        ----------

        g: tohu generator
            The generator to be buffered. The items produced by `g` must be tuples.
        maxbuffer: integer
            Maximum number of items to be buffered.
        """
        assert isinstance(g, TupleGenerator)
        self.g = g
        self.maxbuffer = maxbuffer
        self._reset_queues()

    def _spawn(self):
        return BufferedTuple(self.g._spawn(), maxbuffer=self.maxbuffer)

    def _reset_queues(self):
        self._queues = [deque(maxlen=self.maxbuffer) for _ in range(self.g.tuple_len)]

    def _refill(self):
        item = next(self.g)
        for x, queue in zip(item, self._queues):
            queue.append(x)

    def reset(self, seed):
        self.g.reset(seed)
        self._reset_queues()

    def next_nth(self, n):
        if len(self._queues[n]) == 0:
            self._refill()
        return self._queues[n].popleft()


def Split(g, *, maxbuffer=10):
    """
    Split a tuple generator into individual generators.

    Parameters
    ----------
    g: tohu generator
        The generator to be split. The items produced by `g` must be tuples.
    maxbuffer: integer
        Maximum number of items produced by `g` that will be buffered.
    """
    assert isinstance(g, TupleGenerator)

    g_buffered = BufferedTuple(g, maxbuffer=maxbuffer)

    class NthBuffered(BaseGenerator):
        def __init__(self, g, idx):
            self.g = g
            self.idx = idx

        def __next__(self):
            return self.g.next_nth(self.idx)

        def _spawn(self):
            return NthBuffered(self.g._spawn(), self.idx)

        def reset(self, seed):
            self.g.reset(seed)

    return tuple(NthBuffered(g_buffered, i) for i in range(g.tuple_len))


class Zip(TupleGenerator):
    """
    Create a generator which produces tuples that are
    combined from the elements produced by multiple
    individual generators.
    """

    def __init__(self, *generators, seed=None):
        self._generators = [g._spawn() for g in generators]
        self.reset(seed)

    def __next__(self):
        return tuple(next(g) for g in self._generators)

    def reset(self, seed):
        for g in self._generators:
            g.reset(seed)


def Geolocation():
    """
    Return a pair (Lon, Lat) of iterators producing.
    """
    return Split(GeolocationPair())


from .generators import *
from .custom_generator import CustomGenerator
from .utils import *
from . import generators
from . import custom_generator
from . import utils

__all__ = generators.__all__ + custom_generator.__all__ + utils.__all__

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



from mako.template import Template
from .csv_formatter import CSVFormatter

__all__ = ['ItemCollection']


def wrap_field_for_mako(fld):
    """
    Helper function which wraps a field name such as 'foo'
    into the mako expression '${item.foo}'.
    """
    return "${item." + fld + "}"


def make_mako_template(fields, *, sep=",", line_separator="\n"):
    """
    Helper function. Returns a mako Template which renders
    an individual item to a string.
    """
    field_strings = [wrap_field_for_mako(x) for x in fields.values()]
    return Template(sep.join(field_strings) + line_separator)


class ItemCollection:
    """
    Class representing a collection of items.
    """

    def __init__(self, items, N):
        self.items = items
        self.N = N

    def __len__(self):
        return self.N

    def __repr__(self):
        return "<ItemCollection of length {}>".format(len(self))

    def __iter__(self):
        return iter(self.items)

    def to_csv(self, filename, *, fmt_str=None, fields=None, sep=",", header=None):
        """
        Write item collection to CSV file.

        Parameters
        ----------
        filename: string
            Name of the output CSV file.
        fields: dict
            Dictionary of the form {<column_name>: <field_name>, ...}
        sep: string
            Field separator. Default: ','
        """

        formatter = CSVFormatter(fmt_str=fmt_str, fields=fields, sep=sep, header=header)
        formatter.write(filename, self.items)



# -*- coding: utf-8 -*-
"""
Generator classes to produce random data with specific properties.
"""

import datetime as dt
import pandas as pd
import re
from collections import namedtuple
from itertools import count, islice
from random import Random
from tqdm import tqdm
from .item_collection import ItemCollection

__all__ = [
    'Integer', 'Constant', 'Float', 'Sequential', 'ChooseFrom', 'CharString', 'DigitString',
    'HashDigest', 'GeolocationPair', 'Timestamp',
]


# Note: It would be better to make this an abstract base class
#    (to enforce the interface in subclasses) rather than
#    raising NotImplementedError for methods that are not
#    provided by subclasses, but somehow this interferes with
#    the metaclass CustomGeneratorMeta below.
#
class BaseGenerator:
    """
    Base class for all of tohu's random generators.
    """

    def __iter__(self):
        return self

    def __next__(self):
        raise NotImplementedError("Class {} does not implement method '__next__'.".format(self.__class__.__name__))

    def reset(self, seed):
        raise NotImplementedError("Class {} does not implement method 'reset'.".format(self.__class__.__name__))

    def generate(self, N, *, seed=None, progressbar=False):
        """
        Return sequence of `N` elements.

        If `seed` is not None, the generator is reset
        using this seed before generating the elements.
        """
        if seed is not None:
            self.reset(seed)
        items = islice(self, N)
        if progressbar:
            items = tqdm(items, total=N)
        return ItemCollection(items, N)

    def _spawn(self):
        """
        This method needs to be implemented by derived classes.
        It should return a new object of the same type as `self`
        which has the same attributes but is otherwise independent.
        """
        raise NotImplementedError("Class {} does not implement method '_spawn'.".format(self.__class__.__name__))


class TupleGenerator(BaseGenerator):
    """
    Abstract base class
    """

    @property
    def tuple_len(self):
        """
        Length of tuples produced by this generator.
        """
        try:
            return self._tuple_len
        except AttributeError:
            raise NotImplementedError("Class {} does not implement attribute 'tuple_len'.".format(self.__class__.__name__))

    @tuple_len.setter
    def tuple_len(self, value):
        self._tuple_len = value


class Constant(BaseGenerator):
    """
    Generator which produces a constant sequence (repeating the same value indefinitely).
    """

    def __init__(self, value):
        """
        Parameters
        ----------
        value:
            The constant value produced by this generator.
        """
        self.value = value

    def _spawn(self):
        return Constant(self.value)

    def reset(self, seed=None):
        pass

    def __next__(self):
        return self.value


class Integer(BaseGenerator):
    """
    Generator which produces random integers k in the range lo <= k <= hi.
    """

    def __init__(self, lo, hi, *, seed=None):
        """
        Parameters
        ----------
        lo: integer
            Lower bound (inclusive).
        hi: integer
            Upper bound (inclusive).
        seed: integer (optional)
            Seed to initialise this random generator.
        """
        self.lo = lo
        self.hi = hi
        self.randgen = Random()
        self.reset(seed)

    def _spawn(self):
        return Integer(self.lo, self.hi)

    def reset(self, seed):
        self.randgen.seed(seed)

    def __next__(self):
        return self.randgen.randint(self.lo, self.hi)


class Float(BaseGenerator):
    """
    Generator which produces random floating point numbers x in the range lo <= x <= hi.
    """

    def __init__(self, lo, hi, *, seed=None):
        """
        Parameters
        ----------
        lo: integer
            Lower bound (inclusive).
        hi: integer
            Upper bound (inclusive).
        seed: integer (optional)
            Seed to initialise this random generator.
        """
        self.lo = lo
        self.hi = hi
        self.randgen = Random()
        self.reset(seed)

    def _spawn(self):
        return Float(self.lo, self.hi)

    def reset(self, seed):
        self.randgen.seed(seed)

    def __next__(self):
        return self.randgen.uniform(self.lo, self.hi)


class Sequential(BaseGenerator):
    """
    Generator which produces a sequence of strings
    of the form:

        "PREFIX001"
        "PREFIX002"
        "PREFIX003"
        ...

    Both the prefix and the number of digits can
    be modified by the user.

    Example:
        >>> s = Sequential(prefix="Foobar_", digits=4)
        >>> next(s)
        Foobar_0001
        >>> next(s)
        Foobar_0002
        >>> next(s)
        Foobar_0003
    """

    def __init__(self, *, prefix, digits):
        """
        Parameters
        ----------
        prefix: string
            Prefix to be appended to generated elements.
        digits: integer
            Number of digits to use for the sequential numbering.
            Any numbers will fewer digits will be zero-padded;
            numbers with more digits are unaffected.
        """
        self.prefix = prefix
        self.digits = digits
        self.fmt_str = self.prefix + '{{:0{digits}}}'.format(digits=digits)
        self.reset()

    def _spawn(self):
        return Sequential(prefix=self.prefix, digits=self.digits)

    def reset(self, seed=None):
        """
        Note that this method supports the `seed` argument (for consistency with other generators),
        but its value is ignored - the generator is simply reset to its initial value.
        """
        self.cnt = count(start=1)

    def __next__(self):
        return self.fmt_str.format(next(self.cnt))


class ChooseFrom(BaseGenerator):
    """
    Generator which produces a sequence of items taken from given a given set of elements.
    """

    def __init__(self, values, *, seed=None):
        """
        Parameters
        ----------
        values: list
            List of options from which to choose elements.
        seed: integer (optional)
            Seed to initialise this random generator.
        """
        self.values = values
        self.idxgen = Integer(lo=0, hi=(len(self.values) - 1))
        self.reset(seed)

    def __next__(self):
        """
        Return random element from the list of values provided during initialisation.
        """
        idx = next(self.idxgen)
        return self.values[idx]

    def _spawn(self):
        return ChooseFrom(values=self.values)

    def reset(self, seed):
        self.idxgen.reset(seed)


class CharString(BaseGenerator):
    """
    Generator which produces a sequence of character strings.
    """

    def __init__(self, *, length, chars, seed=None):
        """
        Parameters
        ----------
        length: integer
            Length of the character strings produced by this generator.
        chars: iterable
            Character set to draw from when generating strings.
        seed: integer (optional)
            Seed to initialise this random generator.
        """
        self.length = length
        self.chars = chars
        self.chargen = ChooseFrom(self.chars)
        self.reset(seed)

    def __next__(self):
        chars = [next(self.chargen) for _ in range(self.length)]
        return ''.join(chars)

    def reset(self, seed):
        self.chargen.reset(seed)


class DigitString(CharString):
    """
    Generator which produces a sequence of strings containing only digits.
    """

    def __init__(self, *, length, seed=None):
        """
        Parameters
        ----------
        length: int
            Length of the digit strings produced by this generator.
        seed: integer (optional)
            Seed to initialise this random generator.
        """
        chars = "0123456789"
        self.length = length
        super().__init__(length=length, chars=chars, seed=seed)

    def _spawn(self):
        return DigitString(length=self.length)


class HashDigest(CharString):
    """
    Generator which produces a sequence of hex strings representing hash digest values.
    """

    def __init__(self, *, length, seed=None):
        """
        Parameters
        ----------
        length: int
            Length of the strings produced by this generator.
        seed: integer (optional)
            Seed to initialise this random generator.
        """
        chars = "0123456789ABCDEF"
        self.length = length
        super().__init__(length=length, chars=chars, seed=seed)

    def _spawn(self):
        return HashDigest(length=self.length)


class GeolocationPair(TupleGenerator):
    """
    Generator which produces a sequence of (lon, lat) coordinates.
    """

    def __init__(self):
        self.lon_gen = Float(-180, 180)
        self.lat_gen = Float(-90, 90)
        self.tuple_len = 2

    def _spawn(self):
        return GeolocationPair()

    def __next__(self):
        return (next(self.lon_gen), next(self.lat_gen))

    def reset(self, seed):
        self.lon_gen.reset(seed)
        self.lat_gen.reset(seed)


def Geolocation():
    return Split(GeolocationPair())


class TimestampError(Exception):
    """
    Custom exception raised to indicate Timestamp errors
    (for example when the end time of the timestamp interval
    is before the start time).
    """


class Timestamp(BaseGenerator):
    """
    Generator which produces a timestamp.
    """

    def __init__(self, *, start=None, end=None, date=None, fmt='%Y-%m-%d %H:%M:%S', uppercase=False):
        """
        Initialise timestamp generator.

        Note that `start` and `end` are both inclusive. They can either
        be full timestamps such as 'YYYY-MM-DD HH:MM:SS', or date strings
        such as 'YYYY-MM-DD'. Note that in the latter case `end` is
        interpreted as as `YYYY-MM-DD 23:59:59`, i.e. the day is counted
        in full.

        Args:
            start (date string):  start time
            end   (date string):  end time
            date (str):           string of the form YYYY-MM-DD. This is an alternative (and mutually exclusive)
                                  to specifying `start` and `end`.
            fmt (str):            formatting string for output (same format as accepted by `datetime.strftime`)
            uppercase (bool):     if True, months are formatted with all uppercase letters (default: False)
        """
        if (date is not None):
            if not (start is None and end is None):
                raise TimestampError("Argument `date` is mutually exclusive with `start` and `end`.")

            self.start = dt.datetime.strptime(date, '%Y-%m-%d')
            self.end = self.start + dt.timedelta(hours=23, minutes=59, seconds=59)
        else:
            if (start is None or end is None):
                raise TimestampError("Either `date` or both `start` and `end` must be provided.")

            try:
                self.start = dt.datetime.strptime(start, '%Y-%m-%d')
            except ValueError:
                self.start = dt.datetime.strptime(start, '%Y-%m-%d %H:%M:%S')

            try:
                self.end = dt.datetime.strptime(end, '%Y-%m-%d %H:%M:%S')
            except ValueError:
                end_date = dt.datetime.strptime(end, '%Y-%m-%d')
                self.end = dt.datetime(end_date.year, end_date.month, end_date.day, 23, 59, 59)

        self.dt = int((self.end - self.start).total_seconds())
        self.fmt = fmt
        self.uppercase = uppercase

        if self.dt < 0:
            raise TimestampError("Start time must be before end time. Got: start_time='{}', end_time='{}'."
                                 "".format(self.start, self.end))

        self.offsetgen = Integer(0, self.dt)

    def _spawn(self):
        return Timestamp(
            start=self.start.strftime('%Y-%m-%d %H:%M:%S'),
            end=self.end.strftime('%Y-%m-%d %H:%M:%S'),
            fmt=self.fmt,
            uppercase=self.uppercase)

    def __next__(self):
        s = (self.start + dt.timedelta(seconds=next(self.offsetgen))).strftime(self.fmt)
        return s.upper() if self.uppercase else s

    def reset(self, seed):
        self.offsetgen.reset(seed)


def _create_attribute_generators(custom_generator):
    """
    Scan `custom_generator` for any attributes (both class-level and instance-level))
    which derive from `tohu.generators.BaseGenerator` and return a dictionary
    of the form {attr_name: fresh_generator} containing a freshly spawned
    generator for each of these attributes.
    """
    dict1 = custom_generator.__class__.__dict__
    dict2 = custom_generator.__dict__
    attrgens = {attr_name: obj._spawn() for attr_name, obj in dict1.items() if isinstance(obj, BaseGenerator)}
    attrgens.update({attr_name: obj._spawn() for attr_name, obj in dict2.items() if isinstance(obj, BaseGenerator)})
    return attrgens


def _get_item_class_name(cg_name):
    """
    Return the item name corresponding to the custom generator name `cg_name`.
    This is the "..." part before "...Generator".

    Examples:
        FoobarGenerator -> Foobar
        QuuxGenerator -> Quux
    """
    return re.match('^(.*)Generator$', cg_name).group(1)


def _create_namedtuple_class(item_cls_name, attrgens):
    item_fields = [name for name in attrgens.keys()]
    return namedtuple(item_cls_name, item_fields)


# class ItemCollection:
#     def __init__(self, items, N):
#         """
#         Parameters:
#         -----------
#         items: iterable
#             Sequence of items in the collection.
#         N: int
#             Number of items in the collection. This needs to be specified explicitly
#             because `items` is typically an iterator.
#         """
#         self.items = list(items)
#         self.N = N
#
#     def __repr__(self):
#         return "<ItemCollection of length {}>".format(len(self))
#
#     def __len__(self):
#         return self.N
#
#     def __getitem__(self, i):
#         return self.items[i]
#
#     def __iter__(self):
#         yield from self.items
#
#     def write(self, filename, *, mode='w', header=None, progressbar=True):
#         """
#         Export items to a file.
#
#         Arguments:
#             filename:     Name of the output file.
#             mode:         How to open the file ('w' = write, 'a' = append)
#             header:       Header line printed at the very beginning (remember to add a newline at the end).
#             progressbar:  Whether to display a progress bar while exporting data.
#         """
#         assert mode in ['w', 'a', 'write', 'append'], "Argument 'mode' must be either 'w'/'write' or 'a'/'append'."
#         assert header is None or isinstance(header, str), "Argument 'header' must be a string."
#
#         with open(filename, mode=mode[0]) as f:
#             if header is not None:
#                 f.write(header)
#
#             rng = range(self.N)
#             if progressbar:
#                 rng = tqdm(rng)
#
#             for _, x in zip(rng, self.items):
#                 f.write(format(x))
#
#     def to_df(self):
#         return pd.DataFrame([pd.Series(item._asdict()) for item in self.items])


# class SeedGenerator:
#     """
#     This class is used in custom generators to create a collection of
#     seeds when reset() is called, so that each of the constituent
#     generators can be re-initialised with a different seed in a
#     reproducible way.
#
#     Note: This is almost identical to the `Integer` class above, but
#     we need a version which does *not* inherit from `BaseGenerator`,
#     otherwise the automatic namedtuple creation in `CustomGeneratorMeta`
#     gets confused.
#     """
#
#     def __init__(self):
#         self.r = Random()
#         self.minval = 0
#         self.maxval = sys.maxsize
#
#     def seed(self, value):
#         self.r.seed(value)
#
#     def __iter__(self):
#         return self
#
#     def __next__(self):
#         return self.r.randint(self.minval, self.maxval)
#
#
# class CustomGeneratorMeta(type):
#     def __new__(metacls, cg_name, bases, clsdict):
#         gen_cls = super(CustomGeneratorMeta, metacls).__new__(metacls, cg_name, bases, clsdict)
#         orig_init = gen_cls.__init__
#
#         def gen_init(self, *args, **kwargs):
#             seed = None
#             if 'seed' in kwargs:
#                 seed = kwargs.pop('seed')
#
#             orig_init(self, *args, **kwargs)
#
#             self._attrgens = _create_attribute_generators(self)
#             item_cls_name = _get_item_class_name(cg_name)
#             self.item_cls = _create_namedtuple_class(item_cls_name, self._attrgens)
#
#             self._seed_generator = SeedGenerator()
#
#             def pprint_item(item):
#                 s = Template(
#                     textwrap.dedent("""
#                         <${cls_name}:
#                         % for fld in fieldnames:
#                             ${fld}: ${getattr(item, fld)}
#                         % endfor
#                         >
#                         """)).render(
#                         cls_name=item_cls_name, fieldnames=self.item_cls._fields, item=item)
#                 print(s)
#
#             # Determine how items produced by this generator should be formatted.
#             if not 'FORMAT_STR' in clsdict:
#                 clsdict['FORMAT_STR'] = ",".join([("${" + fld + "}") for fld in self.item_cls._fields]) + '\n'
#                 #print("[DDD] Case 1: {} ({})".format(clsdict['FORMAT_STR'], self))
#             else:
#                 # FIXME: This line seems to be called *twice* for every CustomGenerator!!!
#                 #print("[DDD] Case 2: {} ){})".format(clsdict['FORMAT_STR'], self))
#                 pass
#
#             self.FORMAT_STR = clsdict['FORMAT_STR']
#
#             def format_obj(obj, fmt):
#                 return self._formatter.render(**obj._asdict())
#
#             self.item_cls.__format__ = format_obj
#             self.item_cls.pprint = pprint_item
#             self.item_cls.format = format_obj
#
#             if seed is not None:
#                 self.reset(seed)
#
#         def gen_get_format_str(self):
#             return self._format_str
#         def gen_set_format_str(self, format_str):
#             self._format_str = format_str
#             self._formatter = Template(self._format_str)
#
#         def gen_cls_next(self):
#             attrs = {name: next(obj) for name, obj in self._attrgens.items()}
#             return self.item_cls(**attrs)
#
#         def gen_reset(self, seed):
#             # Reset the seed generator
#             self._seed_generator.seed(seed)
#
#             # Reset each constituent generator with a new seed
#             # produced by the seed generator.
#             for g, x in zip(self._attrgens.values(), self._seed_generator):
#                 g.reset(x)
#
#         def gen_spawn(self):
#             return self.__class__()
#
#         def gen_export(self, filename, *, N, mode='w', seed=None, header=None, progressbar=True):
#             """
#             Produce `N` elements and write them to the file `f`.
#
#             Arguments:
#                 filename:     Name of the output file.
#                 N:            Number of records to write.
#                 mode:         How to open the file ('w' = write, 'a' = append)
#                 seed:         If given, reset generator with this seed.
#                 header:       Header line printed at the very beginning (remember to add a newline at the end).
#                 progressbar:  Whether to display a progress bar while exporting data.
#             """
#             item_collection = ItemCollection(self.generate(N, seed=seed), N)
#             item_collection.write(filename, mode=mode, header=header, progressbar=progressbar)
#
#         gen_cls.__init__ = gen_init
#         gen_cls.__next__ = gen_cls_next
#         gen_cls.reset = gen_reset
#         gen_cls._spawn = gen_spawn
#         gen_cls.export = gen_export
#         gen_cls.FORMAT_STR = property(gen_get_format_str, gen_set_format_str)
#
#         return gen_cls
#
#
# class CustomGenerator(BaseGenerator, metaclass=CustomGeneratorMeta):
#     """
#     The only purpose of this class is to make defining
#     custom generators easier for the user by writing:
#
#         class FooGenerator(CustomGenerator):
#             # ...
#
#     instead of the more clumsy:
#
#         class FooGenerator(metaclass=CustomGeneratorMeta):
#             # ...
#     """
#
#     def generate(self, N, *, seed=None, progressbar=False):
#         return ItemCollection(super().generate(N, seed=seed, progressbar=progressbar), N)



"""
Tests for the CustomGenerator class.
"""

import pytest
import textwrap
from .context import tohu

from tohu.generators import Constant, Integer, Float, Sequential
from tohu.custom_generator import CustomGenerator


class TestCustomGenerator:

    def setup(self):

        class FoobarGenerator(CustomGenerator):
            a = Integer(lo=1000, hi=9000)
            b = Sequential(prefix="foo_", digits=3)

        class QuuxGenerator(CustomGenerator):
            c = Constant("Hello")
            d = Sequential(prefix="quux_", digits=2)
            e = Integer(lo=3000, hi=6000)

        self.gen_foo = FoobarGenerator(seed=99999)
        self.gen_quux = QuuxGenerator(seed=12345)

    def test_custom_generator_produces_objects_of_the_expected_class(self):
        """
        Test that generators FoobarGenerator and QuuxGenerator produce items
        that are instances of a classes named Foobar and Quux, respectively.
        """

        item1 = next(self.gen_foo)
        item2 = next(self.gen_quux)

        assert item1.__class__.__name__ == "Foobar"
        assert item2.__class__.__name__ == "Quux"

    def test_custom_generator_produces_namedtuples_with_expected_field_values(self):
        """
        Test that FoobarGenerator and QuuxGenerator produce namedtuples with fields
        produced by their respective generators.
        """

        item1 = next(self.gen_foo)
        item2 = next(self.gen_foo)

        item3 = next(self.gen_quux)
        item4 = next(self.gen_quux)

        assert item1 == (2184, "foo_001")
        assert item2 == (8875, "foo_002")

        assert item3 == ("Hello", "quux_01", 4001)
        assert item4 == ("Hello", "quux_02", 5032)

    def test_fields_can_be_defined_at_classlevel_and_in_init(self):
        """
        Test that fields can be defined at classlevel an in __init__() of custom generators.
        """
        class QuuxGenerator(CustomGenerator):
            x = Integer(lo=400, hi=499)

            def __init__(self, z_min, z_max):
                self.y = Float(lo=z_min, hi=z_max)

        g1 = QuuxGenerator(z_min=2.0, z_max=3.0)
        g2 = QuuxGenerator(z_min=5.0, z_max=6.0)
        g1.reset(seed=12345)
        g2.reset(seed=12345)

        item1 = next(g1)
        item2 = next(g2)

        assert str(item1) == "Quux(x=488, y=2.032576355272894)"
        assert str(item2) == "Quux(x=488, y=5.032576355272894)"

    def test_formatting_items_returns_string_with_field_values(self):
        """
        Test that formatting items produced by {Foobar|Quux}Generator
        results in strings that are concatenations of their field values.
        """

        item1 = next(self.gen_foo)
        item2 = next(self.gen_foo)

        item3 = next(self.gen_quux)
        item4 = next(self.gen_quux)

        assert str(item1) == "Foobar(a=2184, b='foo_001')"
        assert str(item2) == "Foobar(a=8875, b='foo_002')"
        assert format(item1) == "2184,foo_001"
        assert format(item2) == "8875,foo_002"

        assert str(item3) == "Quux(c='Hello', d='quux_01', e=4001)"
        assert str(item4) == "Quux(c='Hello', d='quux_02', e=5032)"
        assert format(item3) == "Hello,quux_01,4001"
        assert format(item4) == "Hello,quux_02,5032"

    def test_write_csv_file(self, tmpdir):
        """
        Test that generated items can be written to a CSV file.
        """
        filename1 = tmpdir.join("output1.txt").strpath
        filename2 = tmpdir.join("output2.txt").strpath

        self.gen_foo.to_csv(filename1, N=3, seed=12345, header="# Custom header")
        self.gen_foo.to_csv(filename2, N=3, seed=12345, header="# Custom header")

        csv_expected = textwrap.dedent("""\
            # Custom header
            6649,foo_001
            7170,foo_002
            8552,foo_003
            """)

        assert open(filename1).read() == csv_expected
        assert open(filename2).read() == csv_expected

    def test_csv_fields_can_be_defined_in_class_declaration(self, tmpdir):
        """
        Test that fields for CSV export can be defined via the CSV_FIELDS class attribute.
        """
        class FoobarGenerator(CustomGenerator):
            a = Integer(lo=1000, hi=9000)
            b = Sequential(prefix="foo_", digits=3)
            CSV_FIELDS = {"Column 1": "a=${a}", "Column 2": "b=${b}"}

        filename = tmpdir.join("output.csv").strpath

        g = FoobarGenerator()
        g.to_csv(filename, N=3, seed=12345)

        csv_expected = textwrap.dedent("""\
            #Column 1,Column 2
            a=6649,b=foo_001
            a=7170,b=foo_002
            a=8552,b=foo_003
            """)

        assert open(filename).read() == csv_expected

    def test_csv_header_can_be_defined_in_class_declaration(self, tmpdir):
        """
        Test that CSV header can be defined via the CSV_HEADER class attribute .
        """
        class FoobarGenerator(CustomGenerator):
            a = Integer(lo=1000, hi=9000)
            b = Sequential(prefix="foo_", digits=3)
            CSV_HEADER = "# My custom header line"
            CSV_FIELDS = {"Column 1": "a: ${a}", "Column 2": "b: ${b}"}

        filename = tmpdir.join("output.csv").strpath

        g = FoobarGenerator()
        g.to_csv(filename, N=3, seed=12345)

        csv_expected = textwrap.dedent("""\
            # My custom header line
            a: 6649,b: foo_001
            a: 7170,b: foo_002
            a: 8552,b: foo_003
            """)

        assert open(filename).read() == csv_expected

    def test_csv_format_string_can_be_defined_in_class_declaration(self, tmpdir):
        """
        Test that CSV formatting can be defined via the CSV_FMT_STR class attribute .
        """
        class FoobarGenerator(CustomGenerator):
            a = Integer(lo=1000, hi=9000)
            b = Sequential(prefix="foo_", digits=3)
            CSV_FMT_STR = "a=${a} || b: ${b}"

        filename = tmpdir.join("output.csv").strpath

        g = FoobarGenerator()
        g.to_csv(filename, N=3, seed=12345)

        csv_expected = textwrap.dedent("""\
            a=6649 || b: foo_001
            a=7170 || b: foo_002
            a=8552 || b: foo_003
            """)

        assert open(filename).read() == csv_expected


import pytest
import textwrap
from collections import namedtuple
from tohu.csv_formatter_v1 import CSVFormatterV1


@pytest.fixture
def csvfmt(scope="module"):
    return CSVFormatterV1(
        {"Column 1": "a=${a}",
         "Column 2": "b: ${b}",
         "Column 3": "c has value ${c}"
         })


class TestCSVFormatterV1:

    def setup(self):
        Foobar = namedtuple("Foobar", ["a", "b", "c"])
        self.item1 = Foobar(1234, "quux_01", 9.8765)
        self.item2 = Foobar(5678, "quux_02", 2.2222)
        self.items = [self.item1, self.item2]

    def test_format_single_item(self, csvfmt):
        """
        Test that formatting a single item produces the expected string.
        """
        assert csvfmt.format_item(self.item1) == "a=1234,b: quux_01,c has value 9.8765"
        assert csvfmt.format_item(self.item2) == "a=5678,b: quux_02,c has value 2.2222"

    def test_default_csv_formatting(self, csvfmt):
        """
        Test that formatting multiple items produces the expected CSV string.
        """
        expected_csv = textwrap.dedent("""\
            #Column 1,Column 2,Column 3
            a=1234,b: quux_01,c has value 9.8765
            a=5678,b: quux_02,c has value 2.2222
            """)

        csv = csvfmt.to_csv(self.items)

        assert csv == expected_csv

    def test_csv_formatting_without_header_line(self, csvfmt):
        """
        Test that `header=False` produces the expected CSV string (without header line).
        """
        expected_csv = textwrap.dedent("""\
            a=1234,b: quux_01,c has value 9.8765
            a=5678,b: quux_02,c has value 2.2222
            """)

        csv = csvfmt.to_csv(self.items, header=False)

        assert csv == expected_csv

    def test_csv_formatting_with_custom_header_line(self, csvfmt):
        """
        Test that passing a string as 'header' argument uses this string for the header line.
        """
        expected_csv = textwrap.dedent("""\
            # This is a custom header line
            a=1234,b: quux_01,c has value 9.8765
            a=5678,b: quux_02,c has value 2.2222
            """)

        csv = csvfmt.to_csv(self.items, header="# This is a custom header line")

        assert csv == expected_csv

    def test_custom_separator(self):
        """
        Test that the `sep` argument is used as field separator.
        """
        csvfmt = CSVFormatterV1(
            {"Column 1": "a=${a}",
             "Column 2": "b: ${b}",
             "Column 3": "c has value ${c}"
             },
            sep=" || ")

        expected_csv = textwrap.dedent("""\
            #Column 1 || Column 2 || Column 3
            a=1234 || b: quux_01 || c has value 9.8765
            a=5678 || b: quux_02 || c has value 2.2222
            """)

        csv = csvfmt.to_csv(self.items)

        assert csv == expected_csv

    def test_csv_export_to_file_with_given_name(self, csvfmt, tmpdir):
        """
        Test that passing the `path_or_buf` argument exports to the given file.
        """
        expected_csv = textwrap.dedent("""\
            #Column 1,Column 2,Column 3
            a=1234,b: quux_01,c has value 9.8765
            a=5678,b: quux_02,c has value 2.2222
            """)

        filename = tmpdir.join("output.txt")
        csvfmt.to_csv(self.items, filename)
        assert open(filename).read() == expected_csv

    def test_csv_export_to_file_with_given_filehandle(self, csvfmt, tmpdir):
        """
        Test that passing the `path_or_buf` argument exports to the given file.
        """
        expected_csv = textwrap.dedent("""\
            #Column 1,Column 2,Column 3
            a=1234,b: quux_01,c has value 9.8765
            a=5678,b: quux_02,c has value 2.2222
            """)

        filename = tmpdir.join("output.txt")
        fh = open(filename, 'w')
        csvfmt.to_csv(self.items, fh)
        fh.close()
        assert open(filename, 'r').read() == expected_csv



import versioneer

from setuptools import setup, find_packages
from codecs import open
from os import path
from tohu import __version__

here = path.abspath(path.dirname(__file__))

# Read long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='tohu',
    version=versioneer.get_version(),
    description='Create random data in a controllable way',
    long_description=long_description,
    url='https://github.com/maxalbert/tohu',
    author='Maximilian Albert',
    author_email='maximilian.albert@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(),
    install_requires=['mako', 'pandas', 'tqdm'],
    extras_require={
        'dev': ['ipython', 'jupyter'],
        'test': ['pytest', 'nbval'],
    },
    cmdclass=versioneer.get_cmdclass(),
    )



"""
Tests for the ItemCollection class.
"""

import os
import pytest
import textwrap

from .context import tohu
from tohu.item_collection import ItemCollection
from tohu.custom_generator import CustomGenerator
from tohu.generators import ChooseFrom, HashDigest, Integer, Sequential, Timestamp


class TestItemCollection:

    @pytest.mark.parametrize("items, N", [
        ([42, 23, "Hello", 12, "foobar"], 5),
        (range(100), 100),
    ])
    def test_converting_item_collection_to_list_returns_original_items(self, items, N):
        """
        Converting an ItemCollection to a list returns the original items.
        """
        c = ItemCollection(items, N)
        assert list(c) == list(items)

    def test_length(self):
        c = ItemCollection(["hello", "world", "foobar", "quux"], 4)
        assert len(c) == 4

    def test_write_csv(self, tmpdir):
        filename1 = tmpdir.join("output_without_header.csv").strpath
        filename2 = tmpdir.join("output_with_default_header.csv").strpath
        filename3 = tmpdir.join("output_with_custom_header.csv").strpath

        class FoobarGenerator(CustomGenerator):
            aaa = HashDigest(length=6)
            bbb = Sequential(prefix="Foo_", digits=2)
            ccc = Integer(0, 100)

        class QuuxGenerator(CustomGenerator):
            def __init__(self, foo_items):
                self.foo1 = ChooseFrom(foo_items)
                self.foo2 = ChooseFrom(foo_items)
                self.date_str = Timestamp(start="2006-01-01", end="2017-09-01", fmt='%d-%b-%y')

        foo = FoobarGenerator()
        foo_items = list(foo.generate(10, seed=12345))
        quux = QuuxGenerator(foo_items)

        csv_fields = {
            'Column_1': '${foo1.aaa}',
            'Column_2': '${foo1.bbb}',
            'Column_3': '${foo2.aaa}',
            'Column_4': '${foo2.ccc}',
            'Column_5': '${date_str}',
        }

        assert not os.path.exists(filename1)
        assert not os.path.exists(filename2)
        assert not os.path.exists(filename3)
        quux_items = quux.generate(5, seed=99999)
        quux_items.to_csv(filename1, fields=csv_fields, header=False)
        quux_items = quux.generate(5, seed=99999)
        quux_items.to_csv(filename2, fields=csv_fields, header=True)
        quux_items = quux.generate(5, seed=99999)
        quux_items.to_csv(filename3, fields=csv_fields, header="# This is a custom header line")
        assert os.path.exists(filename1)
        assert os.path.exists(filename2)
        assert os.path.exists(filename3)

        expected_output_without_header = textwrap.dedent("""\
            27E995,Foo_04,27E995,58,25-Sep-10
            27E995,Foo_04,9CB45F,36,12-Aug-11
            9CB45F,Foo_10,297F9D,39,17-Jan-07
            1A2721,Foo_06,B2A827,1,05-Jan-16
            F5C771,Foo_08,8CE1AD,68,09-Mar-12
            """)

        expected_output_with_default_header = \
            ("#Column_1,Column_2,Column_3,Column_4,Column_5\n" +
             expected_output_without_header)

        expected_output_with_custom_header = \
            ("# This is a custom header line\n" +
             expected_output_without_header)

        assert open(filename1).read() == expected_output_without_header
        assert open(filename2).read() == expected_output_with_default_header
        assert open(filename3).read() == expected_output_with_custom_header


import re
import sys
from collections import namedtuple
from random import Random

from .generators import BaseGenerator
from .csv_formatter import CSVFormatter
from .csv_formatter_v1 import CSVFormatterV1

__all__ = ["CustomGenerator"]


def get_item_class_name(generator_class_name):
    """
    Given the name of a generator class (such as "FoobarGenerator),
    return the first part of the name before "Generator", which
    will be used for the namedtuple items produced by this generator.

    Examples:
        FoobarGenerator -> Foobar
        QuuxGenerator   -> Quux
    """
    return re.match('^(.*)Generator$', generator_class_name).group(1)


class SeedGenerator:
    """
    This class is used in custom generators to create a collection of
    seeds when reset() is called, so that each of the constituent
    generators can be re-initialised with a different seed in a
    reproducible way.

    Note: This is almost identical to the `Integer` class above, but
    we need a version which does *not* inherit from `BaseGenerator`,
    otherwise the automatic namedtuple creation in `CustomGeneratorMeta`
    gets confused.
    """

    def __init__(self):
        self.r = Random()
        self.minval = 0
        self.maxval = sys.maxsize

    def seed(self, value):
        self.r.seed(value)

    def __iter__(self):
        return self

    def __next__(self):
        return self.r.randint(self.minval, self.maxval)


class CustomGeneratorMeta(type):
    def __new__(metacls, cg_name, bases, clsdict):
        gen_cls = super(CustomGeneratorMeta, metacls).__new__(metacls, cg_name, bases, clsdict)
        orig_init = gen_cls.__init__

        def gen_init(self, *args, **kwargs):
            seed = kwargs.pop('seed', None)

            # Call original __init__ function to make sure all generator attributes are defined
            orig_init(self, *args, **kwargs)

            clsdict = self.__class__.__dict__
            instdict = self.__dict__
            self.field_gens = {name: gen._spawn() for name, gen in dict(**clsdict, **instdict).items() if isinstance(gen, BaseGenerator)}
            clsname = get_item_class_name(self.__class__.__name__)
            self.item_cls = namedtuple(clsname, self.field_gens.keys())

            self.fmt_dict = {name: "${" + name + "}" for name in self.field_gens.keys()}
            self.csvformatter = CSVFormatterV1(self.fmt_dict)

            self.item_cls.__format__ = lambda item, fmt: self.csvformatter.format_item(item)
            self.seed_generator = SeedGenerator()
            self.reset(seed)

        gen_cls.__init__ = gen_init
        return gen_cls


class CustomGenerator(BaseGenerator, metaclass=CustomGeneratorMeta):

    def reset(self, seed=None):
        """
        Reset generator using the given seed (unless seed is None, in which case this is a no-op).
        """
        # Reset the seed generator
        if seed is not None:
            self.seed_generator.seed(seed)

            # Reset each constituent generator with a new seed
            # produced by the seed generator.
            for g, x in zip(self.field_gens.values(), self.seed_generator):
                g.reset(x)

    def __next__(self):
        field_values = [next(g) for g in self.field_gens.values()]
        return self.item_cls(*field_values)

    def to_csv(self, filename, *, N, seed=None, fields=None, fmt_str=None, header=None):
        """
        Generate N items and write output to a CSV file.

        Parameters
        ----------
        filename: string
            Output filename.
        N: integer
            Number of items to generate.
        seed: integer (optional)
            Seed with which to initialise random generator.

        The remaining arguments `fields`, `fmt_str`, `header`
        are passed on to CSVFormatter.
        """
        if fmt_str is None:
            fmt_str = getattr(self, 'CSV_FMT_STR', None)
        if fields is None and fmt_str is None:
            fields = getattr(self, 'CSV_FIELDS', self.fmt_dict)
        if header is None:
            header = getattr(self, 'CSV_HEADER', None)

        formatter = CSVFormatter(fmt_str=fmt_str, fields=fields, header=header)
        formatter.write(filename, self.generate(N, seed=seed))